﻿Imports System.Collections.ObjectModel
Imports System.IO
Imports System.IO.IsolatedStorage
Imports System.Text
Imports System.Xml.Serialization

	Public Class BookInfo
		Public Sub New()
			Me.Paragraphs = New List(Of ParagraphInfo)()
			Me.Bookmarks = New ObservableCollection(Of Bookmark)()
			Me.Annotations = New ObservableCollection(Of Annotation)()
		End Sub

		' Is called once when BookInfo is first created from Project Gutenberg file 
		Public Sub New(ByVal infoFileName As String, ByVal bookFileName As String, ByVal friendlyTitle As String)
			Me.New()
			Me.InfoFilename = infoFileName
			Me.BookFilename = bookFileName
			Me.FriendlyTitle = friendlyTitle

			GenerateParagraphs(bookFileName)
			Me.Chapters = New ObservableCollection(Of ChapterInfo)()
			GenerateChapters()
		End Sub

		Public Property InfoFilename() As String

		Public Property BookFilename() As String

		Public Property FriendlyTitle() As String

		Public Property PaginationPageSize() As Size

		Public Property PaginationFontFamily() As String

		Public Property PaginationFontSize() As Double

		Public Property CurrentChapterIndex() As Integer

		Public Property CurrentPageIndex() As Integer

		Public Property IsChapterAndPageInvalid() As Boolean

		Public Property CurrentCharacterOffset() As Integer

		Public Property TotalCharacterCount() As Integer

		<XmlIgnore>
		Public Property Paragraphs() As List(Of ParagraphInfo)

		Public Property Chapters() As ObservableCollection(Of ChapterInfo)

		Public Property Bookmarks() As ObservableCollection(Of Bookmark)

		Public Property Annotations() As ObservableCollection(Of Annotation)

		Public Shared Function Load(ByVal infoFilename As String) As BookInfo
			Dim book As BookInfo = Nothing

			Using isoStorage As IsolatedStorageFile = IsolatedStorageFile.GetUserStoreForApplication()
				If isoStorage.FileExists(infoFilename) Then
					Using stream As IsolatedStorageFileStream = isoStorage.OpenFile(infoFilename, FileMode.Open)
						Dim serializer As New XmlSerializer(GetType(BookInfo))
						book = TryCast(serializer.Deserialize(stream), BookInfo)
					End Using
				End If
			End Using

			If book IsNot Nothing Then
				book.GenerateParagraphs(book.BookFilename)
			End If

			Return book
		End Function

		Public Sub Save()
			Using isoStorage As IsolatedStorageFile = IsolatedStorageFile.GetUserStoreForApplication()
				Using stream As IsolatedStorageFileStream = isoStorage.CreateFile(Me.InfoFilename)
					Dim serializer As New XmlSerializer(GetType(BookInfo))
					serializer.Serialize(stream, Me)
				End Using
			End Using
		End Sub

		Private Sub GenerateParagraphs(ByVal bookFilename As String)
			Using isoStorage As IsolatedStorageFile = IsolatedStorageFile.GetUserStoreForApplication()
				Using stream As IsolatedStorageFileStream = isoStorage.OpenFile(bookFilename, FileMode.Open)
					GenerateParagraphs(stream)
				End Using
			End Using
		End Sub

		Private Sub GenerateParagraphs(ByVal stream As Stream)
			Dim accumulatedCharacterCount As Integer = 0

			Using reader As New StreamReader(stream)
				Dim line As String = Nothing
				Dim prevLine As String = Nothing
				Dim prevIndent As Integer = 0
				Dim paragraphBuilder As New StringBuilder()

				line = reader.ReadLine()
				Do While Nothing IsNot line
					' Trim line of all space and find the line indent
					line = line.TrimEnd(" "c)
					Dim indent As Integer = line.Length
					line = line.TrimStart(" "c)
					indent -= line.Length

					' A new paragraph is always implied with a line of zero length
					Dim newParagraph As Boolean = line.Length = 0

					If (Not newParagraph) AndAlso paragraphBuilder.Length > 0 AndAlso prevLine IsNot Nothing Then
						' This is the case for a list ("contents" for example)
						'  where each line has a different indent
						If indent <> prevIndent Then
							newParagraph = True

						' But some indented text needs to be concatenated (epistoles, for example)
						ElseIf indent <> 0 Then
							Dim firstWord As String = line
							Dim spaceIndex As Integer = line.IndexOf(" "c)

							If spaceIndex <> -1 Then
								firstWord = firstWord.Substring(0, spaceIndex)
							End If

							' This works for some books, but not for "Phineas Finn" where
							'  at least the first letter has shorter lines than normal
							' if (indent + prevLine.Length + 1 + firstWord.Length <= 70)
                        '     newParagraph = True
						End If
					End If

					' Check for a possible paragraph termination
					If newParagraph Then
						If paragraphBuilder.Length > 0 Then
							SubstituteCharacters(paragraphBuilder)

							' Create a new ParagraphInfo for the paragraph
							Dim paragraph As ParagraphInfo = New ParagraphInfo With {.Indent = prevIndent, .Text = paragraphBuilder.ToString(), .AccumulatedCharacterCount = accumulatedCharacterCount}
							Me.Paragraphs.Add(paragraph)
							accumulatedCharacterCount += paragraph.Text.Length

							' Clear the StringBuilder
							paragraphBuilder.Remove(0, paragraphBuilder.Length)
						End If

						If line.Length > 0 Then
							' Append the line to the paragraph
							paragraphBuilder.Append(line)
						Else
							' Add the current line that terminates the paragraph
							Dim emptyParagraph As ParagraphInfo = New ParagraphInfo With {.Indent = 0, .Text = line}
							Me.Paragraphs.Add(emptyParagraph)
						End If
					Else
						' Insert a space between appended lines assembled into paragraph
						If paragraphBuilder.Length > 0 Then
							paragraphBuilder.Append(" ")
						End If

						paragraphBuilder.Append(line)
					End If
					prevLine = line
					prevIndent = indent
					line = reader.ReadLine()
				Loop

				' Add the last paragraph if there's one in progress
				If paragraphBuilder.Length > 0 Then
					SubstituteCharacters(paragraphBuilder)

					Dim paragraph As ParagraphInfo = New ParagraphInfo With {.Indent = prevIndent, .Text = paragraphBuilder.ToString(), .AccumulatedCharacterCount = accumulatedCharacterCount}
					Me.Paragraphs.Add(paragraph)
					accumulatedCharacterCount += paragraph.Text.Length
				End If
			End Using
		End Sub

		' Does fancy quotes and em-dashes
		Private Sub SubstituteCharacters(ByVal paragraphBuilder As StringBuilder)
			' Replace boring double quotes with fancy quotes
			Dim isCloseQuote As Boolean = False

        For i = 0 To paragraphBuilder.Length - 1
            If paragraphBuilder(i) = """"c Then
                paragraphBuilder(i) = If(isCloseQuote, ChrW(&H201D), ChrW(&H201C))
                isCloseQuote = isCloseQuote Xor True
            End If
        Next i

			' Replace double dashes with em dashes separated by spaces
			'  (but only when they're surrounded by non-dashes)
			Dim index As Integer = 0

			Do While index < paragraphBuilder.Length - 1
				If paragraphBuilder(index) = "-"c AndAlso paragraphBuilder(index + 1) = "-"c AndAlso (index = 0 OrElse paragraphBuilder(index - 1) <> "-"c) AndAlso (index = paragraphBuilder.Length - 2 OrElse paragraphBuilder(index + 2) <> "-"c) Then ' there isn't a later dash -  there isn't an earlier dash -  next one is dash -  current is dash
					If index = 0 AndAlso paragraphBuilder.Length = 2 Then ' double-dash appears alone
						paragraphBuilder.Remove(index, 2)
						paragraphBuilder.Insert(index, ChrW(&H2014).ToString())
					ElseIf index = 0 Then ' double-dash at beginnning of paragraph
						paragraphBuilder.Remove(index, 2)
						paragraphBuilder.Insert(index, ChrW(&H2014).ToString() & " ")
						index += 1
					ElseIf index >= paragraphBuilder.Length - 3 Then ' double-dash not followed by at least two characters
						paragraphBuilder.Remove(index, 2)
						paragraphBuilder.Insert(index, " " & ChrW(&H2014).ToString())
						index += 1
					Else
						' Double quote-followed by em-dash ("Bleak House")
						If index > 0 AndAlso paragraphBuilder(index - 1) = ChrW(&H201C) Then
							paragraphBuilder.Remove(index, 2)
							paragraphBuilder.Insert(index, ChrW(&H2014).ToString() & " ")
							index += 1
						' em-dash followed by double-quote ("Ditto")
						ElseIf index < paragraphBuilder.Length - 1 AndAlso paragraphBuilder(index + 2) = ChrW(&H201D) Then
							paragraphBuilder.Remove(index, 2)
							paragraphBuilder.Insert(index, " " & ChrW(&H2014).ToString())
							index += 1
						Else
							paragraphBuilder.Remove(index, 2)
							paragraphBuilder.Insert(index, " " & ChrW(&H2014).ToString() & " ")
							index += 2
						End If
					End If
				End If

				index += 1
			Loop
		End Sub

		Private Sub GenerateChapters()
			' Set so the beginning of the book becomes the first chapter
			Dim consecutiveBlankLineCount As Integer = 10
			Dim accumulatedCharacterCount As Integer = 0
			Dim chapterInfo As ChapterInfo = Nothing

        For paragraphIndex = 0 To Me.Paragraphs.Count - 1
            If Me.Paragraphs(paragraphIndex).Text.Length = 0 Then
                consecutiveBlankLineCount += 1
            Else
                If consecutiveBlankLineCount > 2 Then
                    Dim pageInfo As PageInfo = New PageInfo With {.ParagraphIndex = paragraphIndex, .CharacterIndex = 0, .AccumulatedCharacterCount = accumulatedCharacterCount}

                    chapterInfo = New ChapterInfo With {.Title = If(Me.Paragraphs(paragraphIndex).Text.Length < 100, Me.Paragraphs(paragraphIndex).Text, Me.Paragraphs(paragraphIndex).Text.Substring(0, 100)), .FirstParagraph = paragraphIndex, .ParagraphCount = 1}
                    chapterInfo.Pages.Add(pageInfo)
                    Me.Chapters.Add(chapterInfo)
                Else
                    ' Accumulate number of paragraphs in the chapter
                    chapterInfo.ParagraphCount += 1 + consecutiveBlankLineCount
                End If
                consecutiveBlankLineCount = 0
                accumulatedCharacterCount += Me.Paragraphs(paragraphIndex).Text.Length
            End If
        Next paragraphIndex

			' Flag the last chapter
			Me.Chapters(Me.Chapters.Count - 1).IsLastChapter = True
			Me.TotalCharacterCount = accumulatedCharacterCount
		End Sub
	End Class
